/* lisp.h   by Michael Thorpe   2005-07-12 */

#ifndef OBJ_H
#define OBJ_H

/* Only necessary for stream objects: */
#include <stdio.h>

typedef struct obj obj;
typedef struct objtype objtype;
typedef struct scope scope;

struct objtype {
	char *name;
	obj *nameobj;
	obj *(*tostring)(obj *);
	void (*dropobj)(obj *); /* Called when object is to be discarded */
	obj *(*evaluate)(scope *,obj *);
	obj *(*applyfunc)(scope *,obj *,obj *);
};

union rawfuncptr {
	obj *(*ns0)();
	obj *(*ns1)(obj *);
	obj *(*ns2)(obj *,obj *);
	obj *(*ns3)(obj *,obj *,obj *);
	obj *(*s0)(scope *);
	obj *(*s1)(scope *,obj *);
	obj *(*s2)(scope *,obj *,obj *);
	obj *(*s3)(scope *,obj *,obj *,obj *);
};

struct obj {
	objtype *type;
	unsigned long refcount;
	union {
		struct c {
			obj *a;
			obj *d;
		} c;
		signed long i;
		double f;
		struct func {
			obj *args;
			obj *prog;
		} func;
		obj *o;
		struct rawfunc {
			unsigned int numargs:2;
			unsigned int wantscope:1;
			union rawfuncptr rawfuncptr;
		} rawfunc;
		struct s {
			char *data;
			unsigned long len;
		} s;
		struct stream {
			unsigned int readable:1;
			unsigned int writable:1;
			unsigned int closable:1;
			FILE *f;
		} stream;
		void *p;
	} value;
#ifdef DEBUG
	obj *prev,*next;
#endif
};

extern objtype
	cons_objtype,
	float_objtype,
	func_objtype,
	integer_objtype,
	nil_objtype,
	quote_objtype,
	rawfunc_objtype,
	stream_objtype,
	string_objtype,
	word_objtype;

struct scope {
	unsigned long refcount;
	scope *parent;
	struct entry *firstentry;
};

/* builtin.c */
int builtin_init();

/* cons.c */
#define CAR(o) ((o)->value.c.a)
#define CDR(o) ((o)->value.c.d)
#define CADR(o) CAR(CDR(o))
#define CADDR(o) CAR(CDR(CDR(o)))
#define CDDR(o) CDR(CDR(o))
obj *newlist(unsigned long length);

/* integer.c */
extern obj trueobj;

/* nil.c */
extern obj nilobj;

/* quote.c */
obj *quoteobj(obj *o);
int quote_init();

/* reader.c */
obj *read_object(obj *stream);
int reader_init();

/* scope.c */
void drop_scope(scope *scope);
#define decscoperef(s) do{if(!--(s)->refcount)drop_scope(s);}while(0)
obj *define(scope *scope,obj *name,obj *value);
obj *define_local(scope *scope,obj *name,obj *value);
#define incscoperef(s) ((s)->refcount++,(s))
obj *lookup(scope *scope,obj *name);
scope *newscope(scope *parent);
int store_builtin(unsigned int numargs,unsigned int wantscope,const char *name,obj *(*rawfunc)());
#define store_builtin_ns0(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)()))store_builtin)(0,0,name,func)
#define store_builtin_ns1(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(obj *)))store_builtin)(1,0,name,func)
#define store_builtin_ns2(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(obj *,obj *)))store_builtin)(2,0,name,func)
#define store_builtin_ns3(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(obj *,obj *,obj *)))store_builtin)(3,0,name,func)
#define store_builtin_s0(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(scope *)))store_builtin)(0,1,name,func)
#define store_builtin_s1(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(scope *,obj *)))store_builtin)(1,1,name,func)
#define store_builtin_s2(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(scope *,obj *,obj *)))store_builtin)(2,1,name,func)
#define store_builtin_s3(name,func) ((int (*)(unsigned int numargs,unsigned int wantscope,char *,obj *(*)(scope *,obj *,obj *,obj *)))store_builtin)(3,1,name,func)
int store_object(const char *name,obj *o);
extern scope *globalscope;
#ifdef DEBUG
extern unsigned long total_scopes;
#endif
int scope_init();

/* stream.c */
extern obj *eoferror,*ioerror,*stderrobj;
/*
 * These next three functions don't decref stream; return values are:
 *    -1 = end-of-file
 *    -2 = io-error
 * other = character read/written
 */
obj *writeobj(obj *stream,obj *o);
int real_sgetc(obj *stream);
int real_speekc(obj *stream);
int real_sputc(obj *stream,int c);
int stream_init();

/* string.c */
obj *newstring(const char *data,unsigned long len);
int string_init();

/* throw.c */
extern obj
	*exception,
	*exceptionvalue;
obj *throw(obj *type,obj *value);
obj *throwaccesserror(obj *value);
obj *throwargumenterror(obj *value);
obj *throwoom();
obj *throwrangecheck(obj *value);
obj *throwtypeerror(obj *value);
int throw_init();

/* util.c */
#ifdef DEBUG
#define decref(o) real_decref((o),__FILE__,__LINE__)
void real_decref(obj *o,const char *file,int line);
#define incref(o) real_incref((o),__FILE__,__LINE__)
obj *real_incref(obj *o,const char *file,int line);
extern unsigned long total_objects,known_objects;
#define INC_KNOWN_OBJECTS(n) (known_objects+=(n))
#else
#define decref(o) do{if(!--(o)->refcount)drop_object(o);}while(0)
#define incref(o) ((o)->refcount++,(o))
#define INC_KNOWN_OBJECTS(n) do{}while(0)
#endif
void drop_object(obj *o);
obj *evaluate(scope *scope,obj *o);
obj *newobj(objtype *type);
int register_objtype(objtype *type);
#ifdef DEBUG
void show_remaining_objects();
#endif
int util_init();

/* word.c */
obj *newword(const char *data,unsigned long len);
int real_eqv(obj *o1,obj *o2);

#endif
