/* reader.c   by Michael Thorpe   2005-07-15 */

#include <ctype.h>
#include <stdlib.h>
#include "lisp.h"

static obj *close_paren,*dot_word,*list_word,*no_object,*reader_error,*unquote_word;

static obj *macro_tick(obj *stream) {
	obj *o;

	o=read_object(incref(stream));
	if(!o) {
		if(real_eqv(exception,eoferror))
			return(throw(incref(reader_error),stream));
		decref(stream);
		return(0);
	}
	decref(stream);
	return(quoteobj(o));
}

static obj *submacro_dot(obj *stream) {
	obj *tmp1,*tmp2;

	tmp1=read_object(incref(stream));
	if(!tmp1) {
		if(real_eqv(exception,eoferror))
			return(throw(incref(reader_error),stream));
		decref(stream);
		return(0);
	}
	if(!real_eqv(tmp1,close_paren)) {
		tmp2=read_object(incref(stream));
		if(!tmp2) {
			decref(tmp1);
			if(real_eqv(exception,eoferror))
				return(throw(incref(reader_error),stream));
			decref(stream);
			return(0);
		}
		if(real_eqv(tmp2,close_paren)) {
			decref(stream);
			decref(tmp2);
			return(tmp1);
		}
		decref(tmp2);
	}
	decref(tmp1);
	return(throw(incref(reader_error),stream));
}

static obj *macro_openparen(obj *stream) {
	obj *start=0,*cur=0;
	obj *tmp;

	while(1) {
		tmp=read_object(incref(stream));
		if(!tmp) {
			if(cur) {
				CDR(cur)=incref(&nilobj);
				decref(start);
			}
			if(real_eqv(exception,eoferror))
				return(throw(incref(reader_error),stream));
			decref(stream);
			return(0);
		}
		if(real_eqv(tmp,close_paren)) {
			decref(tmp);
			break;
		}
		if(real_eqv(tmp,dot_word)) {
			decref(tmp);
			if(!cur)
				return(throw(incref(reader_error),stream));
			CDR(cur)=submacro_dot(incref(stream));
			if(!CDR(cur)) {
				CDR(cur)=incref(&nilobj);
				decref(start);
				decref(stream);
				return(0);
			}
			decref(stream);
			return(start);
		}
		if(start) {
			CDR(cur)=newobj(&cons_objtype);
			if(!CDR(cur)) {
				CDR(cur)=incref(&nilobj);
				decref(start);
				decref(stream);
				return(0);
			}
			cur=CDR(cur);
		} else {
			cur=start=newobj(&cons_objtype);
			if(!start) {
				decref(stream);
				return(0);
			}
		}
		CAR(cur)=tmp;
	}
	decref(stream);
	if(cur)
		CDR(cur)=incref(&nilobj);
	else
		start=incref(&nilobj);
	return(start);
}

static obj *macro_closeparen(obj *stream) {
	decref(stream);
	return(incref(close_paren));
}

static obj *macro_comma(obj *stream) {
	obj *cons,*foo;

	foo=read_object(stream);
	if(!foo) {
		if(real_eqv(exception,eoferror))
			return(throw(incref(reader_error),stream));
		decref(stream);
		return(0);
	}
	decref(stream);
	cons=newlist(2);
	if(!cons) {
		decref(foo);
		return(0);
	}
	decref(CAR(cons));
	CAR(cons)=incref(unquote_word);
	decref(CADR(cons));
	CADR(cons)=foo;
	return(cons);
}

static obj *macro_semicolon(obj *stream) {
	int c;

	while(1) {
		c=real_sgetc(stream);
		switch(c) {
		case -1:
			return(throw(incref(eoferror),stream));
		case -2:
			return(throw(incref(ioerror),stream));
		case '\n':
		case '\r':
			decref(stream);
			return(incref(no_object));
		}
	}
}

static obj *quasiquote(obj *arg) {
	obj *new,*old,*result,*tmp;

	if(arg->type==&nil_objtype)
		return(arg);
	if(arg->type != &cons_objtype) {
		result=newobj(&quote_objtype);
		if(!result) {
			decref(arg);
			return(0);
		}
		result->value.o=arg;
		return(result);
	}
	if(CAR(arg)==unquote_word && CDR(arg)->type==&cons_objtype && CDDR(arg)==&nilobj) {
		result=incref(CADR(arg));
		decref(arg);
		return(result);
	}
	result=newlist(2);
	if(!result) {
		decref(arg);
		return(0);
	}
	decref(CAR(result));
	CAR(result)=incref(list_word);
	new=CDR(result);
	old=arg;
	while(1) {
/*
 * At this point:
 *   new is !0
 *   old and new are pointing to the same place on their respective lists
 */
		tmp=quasiquote(incref(CAR(old)));
		if(!tmp) {
			decref(result);
			decref(arg);
			return(0);
		}
		decref(CAR(new));
		CAR(new)=tmp;
		old=CDR(old);
		if(old->type != &cons_objtype)
			break;
		tmp=newlist(1);
		if(!tmp) {
			decref(result);
			decref(arg);
			return(0);
		}
		decref(CDR(new));
		CDR(new)=tmp;
		new=CDR(new);
	}
	tmp=quasiquote(incref(old));
	if(!tmp) {
		decref(result);
		decref(arg);
		return(0);
	}
	decref(CDR(new));
	CDR(new)=tmp;
	decref(arg);
	return(result);
}

static obj *macro_backtick(obj *stream) {
	obj *foo;

	foo=read_object(stream);
	if(!foo)
		return(0);
	return(quasiquote(foo));
}

#define MAX_STRING 8192
static char string_buf[MAX_STRING+1];

static obj *macro_verticalbar(obj *stream) {
	int c,i=0,seenbackslash=0;

	while(1) {
		c=real_sgetc(stream);
		if(-1==c)
			return(throw(incref(eoferror),stream));
		if(-2==c)
			return(throw(incref(ioerror),stream));
		if('\\'==c) {
			seenbackslash=!seenbackslash;
			if(seenbackslash)
				continue;
		} else if('|'==c) {
			if(!seenbackslash) {
				decref(stream);
				return(newstring(string_buf,i));
			}
			seenbackslash=0;
		}
		string_buf[i++]=c;
		if(MAX_STRING<i)
			return(throw(incref(reader_error),stream));
	}
}

#define MAX_TOKEN 128
static char reader_buf[MAX_TOKEN+1];

#define ES_NORMAL 0
#define ES_SINGLE 1
#define ES_MULTIPLE 2
#define ES_MULTIPLESINGLE 3

/* This should be static */
obj *read_object(obj *stream) {
	obj *o;
	int c,buflen=0,isnum=0;
	int escapestate=ES_NORMAL;
	char *tmp;

	if(stream->type != &stream_objtype)
		return(throwtypeerror(stream));
	while(1) {
		c=real_speekc(stream);
		if(c==-1) {
			if(escapestate==ES_NORMAL && buflen)
				break;
			return(throw(incref(eoferror),stream));
		}
		if(!isascii(c))
			return(throw(incref(reader_error),stream));
		if(isspace(c)) {
			if(escapestate==ES_NORMAL) {
				if(buflen)
					break;
				else {
					real_sgetc(stream);
					continue;
				}
			}
		}
		/* Terminating macro characters */
		if(escapestate==ES_NORMAL && (c=='\'' || c=='(' || c==')' || c==',' || c==';' || c=='`' || c=='|')) {
			if(buflen)
				break;
			real_sgetc(stream);
			incref(stream);
			switch(c) {
			case '\'': o=macro_tick(stream); break;
			case '(': o=macro_openparen(stream); break;
			case ')': o=macro_closeparen(stream); break;
			case ',': o=macro_comma(stream); break;
			case ';': o=macro_semicolon(stream); break;
			case '`': o=macro_backtick(stream); break;
			case '|': o=macro_verticalbar(stream); break;
			default: o=0; /* This is just to shut gcc up */
			}
			if(o && real_eqv(o,no_object)) {
				decref(o);
				continue;
			}
			decref(stream);
			return(o);
		}
		real_sgetc(stream);
		if(c=='\\' && !(escapestate&ES_SINGLE)) {
			escapestate |= ES_SINGLE;
			continue;
		}
		if(c=='"' && !(escapestate&ES_SINGLE)) {
			escapestate ^= ES_MULTIPLE;
			continue;
		}
/* Is everything else a constituent character? */
		if((!buflen || (buflen==1 && reader_buf[0]=='-')) && isdigit(c))
			isnum=1;
		if(isnum && !isdigit(c))
			isnum=0;
		reader_buf[buflen++]=c;
		if(buflen==MAX_TOKEN+1)
			return(throw(incref(reader_error),stream));
		escapestate &= ~ES_SINGLE;
	}
	decref(stream);
	if(isnum) {
		o=newobj(&integer_objtype);
		if(!o)
			return(0);
		reader_buf[buflen]='\0';
		o->value.i=strtol(reader_buf,&tmp,10);
		if(buflen != tmp-reader_buf)
			return(throwrangecheck(o));
		return(o);
	}
	return(newword(reader_buf,buflen));
}

int reader_init() {
	if(!(close_paren=newword(")",1))
	|| !(dot_word=newword(".",1))
	|| !(list_word=newword("list",4))
	|| !(no_object=newword("no-object",9))
	|| store_builtin_ns1("read-object",read_object)
	|| !(reader_error=newword("reader_error",12))
	|| !(unquote_word=newword("unquote",7)))
		return(1);
	INC_KNOWN_OBJECTS(6);
	return(0);
}

